#ifndef __AMATH_H
#define __AMATH_H

/*
** Fixed point manipulation routines
*/
#define _FloatScaler 65536.0
#define _FixShift    16

typedef long AstFixedpoint_t;

inline AstFixedpoint_t AstConvertToFix( const double a )
{
   return ( AstFixedpoint_t ) ( a * _FloatScaler );
}

inline AstFixedpoint_t AstConvertToFix( const float a )
{
   return ( AstFixedpoint_t ) ( a * _FloatScaler );
}

inline AstFixedpoint_t AstConvertToFix( const long a )
{
   return ( AstFixedpoint_t ) ( a << _FixShift );
}

inline AstFixedpoint_t AstConvertToFix( const short int a )
{
   return ( AstFixedpoint_t ) ( ( ( long ) a ) << _FixShift );
}

#ifdef __BORLANDC__
inline AstFixedpoint_t AstConvertToFix( const int a )
{
   return AstConvertToFix( ( short ) a );
}
#else
inline AstFixedpoint_t AstConvertToFix( const int a )
{
   return AstConvertToFix( ( long ) a );
}
#endif

inline float AstConvertToFloat( const AstFixedpoint_t a )
{
   return ( ( a ) / _FloatScaler );
}

inline short AstConvertToInt( const AstFixedpoint_t a )
{
   return ( short ) ( ( a ) >> _FixShift );
}

extern AstFixedpoint_t AstFixMulFix( AstFixedpoint_t a, AstFixedpoint_t b );
extern AstFixedpoint_t AstFixDivFix( AstFixedpoint_t a, AstFixedpoint_t b );
#ifndef __BORLANDC__
#  pragma aux AstFixDivFix parm [eax] [ebx] modify [edx] value [eax] = \
        "cdq" \
        "shld edx, eax, 16" \
        "sal eax,16" \
        "idiv ebx";
#  pragma aux AstFixMulFix parm [eax] [ebx] modify [edx] value [eax] = \
        "imul ebx" \
        "shrd eax,edx,16";
#endif

/*
** Vector and Matrix Definitions
*/
struct AstVector;
class  AstMatrix3x3;
class  AstMatrix4x3;

struct AstVector
{
  AstFixedpoint_t X, Y, Z;

  void operator *= ( AstFixedpoint_t scaler );
  void operator *= ( const AstMatrix3x3 &m );

  void operator += ( const AstVector &v )
  {
    X += v.X; Y += v.Y; Z += v.Z;
  }
  void operator -= ( const AstVector &v )
  {
    X -= v.X; Y -= v.Y; Z -= v.Z;
  }
  void SetZero( void )
  {
    X = Y = Z = 0L;
  }

  AstVector       Cross( const AstVector &v ) const;
  AstFixedpoint_t Dot( const AstVector &v ) const;
  void            Normalize( void );
  AstFixedpoint_t Magnitude( void ) const;
  int             MagnitudeSquared( void ) const;

  AstVector( void ) {}
  AstVector( AstFixedpoint_t x, AstFixedpoint_t y, AstFixedpoint_t z )
  {
     X = x;
     Y = y;
     Z = z;
  }
};

AstVector operator - ( const AstVector &v );


class AstMatrix3x3
{
  protected:
    AstFixedpoint_t data[3][3];

  public:

    void Transpose( void );

    const AstFixedpoint_t * operator[] ( int a ) const { return data[a]; }
    AstFixedpoint_t       * operator[] ( int a ) { return data[a]; }
};

class AstMatrix4x3
{
  protected:
    AstFixedpoint_t data[4][3];

  public:

    void MakeIdentity( void );

    void MakeXTranslation( AstFixedpoint_t amount );
    void MakeYTranslation( AstFixedpoint_t amount );
    void MakeZTranslation( AstFixedpoint_t amount );

    void MakeXRotation( unsigned short amount );
    void MakeYRotation( unsigned short amount );
    void MakeZRotation( unsigned short amount );

    void NormalizeDirectionVectors( void );

    void GetRotationMatrix( AstMatrix3x3 &rm ) const;
    void SetRotationMatrix( const AstMatrix3x3 &rm );
    void SetTranslation( AstFixedpoint_t x, AstFixedpoint_t y, AstFixedpoint_t z )
    {
       data[3][0] = x;
       data[3][1] = y;
       data[3][2] = z;
    }
    void SetTranslation( const AstVector &v )
    {
       SetTranslation( v.X, v.Y, v.Z );
    }

    AstVector GetTranslation( void ) const
    {
       return AstVector( data[3][0], data[3][1], data[3][2] );
    }

    void GetVU( AstVector &v ) const
    {
       v.X = data[0][1];
       v.Y = data[1][1];
       v.Z = data[2][1];
    }
    void GetVR( AstVector &v ) const
    {
       v.X = data[0][0];
       v.Y = data[1][0];
       v.Z = data[2][0];
    }
    void GetVF( AstVector &v ) const
    {
       v.X = data[0][2];
       v.Y = data[1][2];
       v.Z = data[2][2];
    }

    void SetVU( const AstVector &v )
    {
       data[0][1] = v.X;
       data[1][1] = v.Y;
       data[2][1] = v.Z;
    }
    void SetVR( const AstVector &v )
    {
       data[0][0] = v.X;
       data[1][0] = v.Y;
       data[2][0] = v.Z;
    }
    void SetVF( const AstVector &v )
    {
       data[0][2] = v.X;
       data[1][2] = v.Y;
       data[2][2] = v.Z;
    }

    const AstFixedpoint_t * operator[] ( int a ) const { return data[a]; }
    AstFixedpoint_t       * operator[] ( int a ) { return data[a]; }
};

AstMatrix4x3 operator * ( const AstMatrix4x3 &a, const AstMatrix4x3 &b );
AstVector    operator * ( const AstVector &v, const AstMatrix3x3 &m );
AstVector    operator * ( const AstVector &v, const AstMatrix4x3 &m );

inline AstVector operator + ( const AstVector &a, const AstVector &b )
{
   return AstVector( a.X + b.X, a.Y + b.Y, a.Z + b.Z );
}

inline AstVector operator - ( const AstVector &a, const AstVector &b )
{
   return AstVector( a.X - b.X, a.Y - b.Y, a.Z - b.Z );
}

#endif
