#include "ast3d.h"

/*
** AstObject::AstObject( const char *n )
**
** The AstObject class is the primary class used throughout
** AST3D -- it provides the fundamental capabilities that a
** visible object has.  This constructor sets up all of the
** miscellaneous variables used by AstObject.
*/
AstObject::AstObject( const char *n )
{
   if ( n )
     SetName( n );

  local_matrix.MakeIdentity();

  neardef = fardef = curdef = 0;
  neardef_Z = fardef_Z      = 0;

  scratch_vertex_array    = 0;
  scratch_screen_array    = 0;
  scratch_cull_array      = 0;
  scratch_intensity_array = 0;

  bounding_sphere = 0;
  bounding_sphere_squared = 0;

  SetFlag( ofVisible, false );
  SetFlag( ofValid, true );
  SetFlag( ofSolid, false );

  next_in_world = 0;
}

/*
** AstObject::~AstObject()
**
** This is the destructor for the AstObject class.  It's only
** duty is to free up memory tied up by the object's scratch
** arrays.
*/
AstObject::~AstObject()
{
   FreeScratchArrays();
}

/*
** void AstObject::AllocateScratchArrays( void )
**
** During the course of rendering individual objects will often
** need temporary areas to store intermediate values.  This
** function allocates these scratch arrays, which include arrays
** to store transformed vertices, screen vertices, culling data,
** and shading values.
*/
void AstObject::AllocateScratchArrays( void )
{
int nv = neardef->GetNumVertices();
int nt = neardef->GetNumTriangles();

  FreeScratchArrays();
  if ( (scratch_vertex_array = new AstVector[nv]) == 0 )
    AstFatal( "Failure to allocate scratch_vertex_array" );
  if ( ( scratch_screen_array = new AstScreenVertex[nv] ) == 0 )
    AstFatal( "Failure to allocate scratch_screen_array" );
  if ( ( scratch_cull_array = new bool[nt] ) == 0 )
     AstFatal( "Failure to allocate scratch_cull_array" );
  if ( ( scratch_intensity_array = new AstFixedpoint_t[nt] ) == 0 )
     AstFatal( "Failure to allocate scratch_intensity_array" );
}

/*
** void AstObject::CalcExtents( void )
**
** This function calculates an object's bounding sphere based
** on the object's near AstObjectDefinition.
*/
void AstObject::CalcExtents( void )
{
  if ( neardef )
  {
     AstFixedpoint_t farthest = 0L;

     /*
     ** Iterate through the object definition's vertices,
     ** and store the distance of the vertex farthest away
     ** from the object's origin.
     */
     for ( int i = 0; i < neardef->GetNumVertices(); i++ )
     {
        if ( neardef->GetVertex( i ).Magnitude() > farthest )
           farthest = neardef->GetVertex( i ).Magnitude();
     }
     bounding_sphere         = farthest;
     bounding_sphere_squared = AstConvertToInt( AstFixMulFix( farthest, farthest ) );
  }
  else
  {
     bounding_sphere = 0;
     bounding_sphere_squared = 0;
  }
}

/*
** const AstMatrix4x3 *AstObject::CalculateWorldToObjectMatrix()
**
** This is an incredibly important function, responsible for
** coming up with the 4x3 matrix that, when applied to a vector,
** transforms that vector from the world coordinate system into
** this object's local coordinate system.  This is important for
** rendering, since this takes WCS data and transforms it into
** the CCS, and also for various optimizations that do
** calculations in an object's local coordinate system.
*/
const AstMatrix4x3 *AstObject::CalculateWorldToObjectMatrix( void )
{
AstVector    position;
AstMatrix3x3 rm3x3;

   /*
   ** Get object's rotation matrix and translation.
   */
   local_matrix.GetRotationMatrix( rm3x3 );

   /*
   ** Invert the rotation matrix (a transpose is the
   ** the same as an inversion when working with
   ** rotation matrices).
   */
   rm3x3.Transpose();

   /*
   ** Transform its location by the inverted rotation matrix
   */
   position = GetPositionWCS() * rm3x3;

   /*
   ** Create world_to_object_matrix
   */
   world_to_object_matrix.SetRotationMatrix( rm3x3 );
   world_to_object_matrix.SetTranslation( -position );
   world_to_object_matrix.NormalizeDirectionVectors();

   return &world_to_object_matrix;
}

/*
** bool AstObject::CheckProjectedBoundingSphere()
**
** This determines if the object's projected bounding sphere is
** visible within the viewport.
*/
bool AstObject::CheckProjectedBoundingSphere( void )
{
AstScreenVertex projected_location, projected_sphere;
AstVector       bsphere( bounding_sphere,
                         bounding_sphere,
                         location_CCS.Z );
AstVector       scratch = location_CCS;

   /*
   ** This isn't the most elaborate method for dealing with the
   ** "danger zone" of Z values, i.e. that set of Z-values that
   ** may cause overflow/underflow/divide by zero, but it's better
   ** than doing true 3D clipping which would remove a lot of
   ** of the speed gains realized by checking projected bounding
   ** spheres.
   */
   if ( scratch.Z < AstConvertToFix( 1 ) )
      scratch.Z = AstConvertToFix( 1 );

   /*
   ** Project the object position and bounding sphere
   */
   camera.ProjectPoints( &scratch, &projected_location, 1 );
   camera.ProjectPoints( &bsphere, &projected_sphere, 1 );

   if ( projected_location.X+projected_sphere.X< 0 ||
        projected_location.X-projected_sphere.X > SCREEN_WIDTH ||
        projected_location.Y+projected_sphere.Y < 0 ||
        projected_location.Y-projected_sphere.Y > SCREEN_HEIGHT )
   {
      return false;
   }
   return true;
}

/*
** AstFixedpoint_t ComputeIntensity( light, normal )
**
** This function computes the intensity, from 0 to 1, of the
** light hitting a surface with surface normal "normal".  This
** is based on a scaled dot product.
*/
static inline AstFixedpoint_t
ComputeIntensity( const AstVector &light,
                  const AstVector &normal )
{
AstFixedpoint_t val = light.Dot( normal )+AstConvertToFix( 1 );
   return AstFixDivFix( val, AstConvertToFix( 2 ) );
}

/*
** bool CheckPlaneEquation( point, normal, plane constant )
**
** This checks to see if the given point is in front of the
** plane defined by the given normal and plane constant.
*/
static inline bool
CheckPlaneEquation( const AstVector &point,
                    const AstVector &normal,
                    const AstFixedpoint_t pc )
{
   if ( point.Dot( normal ) + pc >= 0 )
      return false;
   else
      return true;
}

/*
** void AstObject::CullAndShade( void )
**
** This routine is responsible for culling out faces that are
** behind the camera and for shading those that are not culled.
*/
void AstObject::CullAndShade( void )
{
AstVector camera_position;
AstVector light_source;
AstMatrix3x3 rm3x3;

   /*
   ** We're going to do our calculations in this object's LCS,
   ** so we need the world_to_object_matrix updated.
   */
   CalculateWorldToObjectMatrix();

   /*
   ** We transform the global light source into object (local)
   ** space by applying the world_to_object_matrix's rotation
   ** matrix to it.  We use only a rotation matrix because
   ** light sources do not possess a translation.
   */
   world_to_object_matrix.GetRotationMatrix( rm3x3 );
   light_source = world.GetLightSource() * rm3x3;

   /*
   ** We transform the camera's position in the WCS into a point
   ** in the object's LCS by applying the world_to_object_matrix
   ** to the camera position.
   */
   camera_position = player->GetPositionWCS() *
                     world_to_object_matrix;

   /*
   ** For each of the triangles, determine if the camera is in
   ** front of or behind the plane of that triangle.  For those
   ** triangles that are not culled out, we calculate an
   ** intensity value for that face used later for shading.
   */

   /*
   ** Note that we do two different loops, one with shading and
   ** one without depending on whether texture mapping is
   ** enabled or not.
   */
   if ( config.texturing ) {
      for ( int i = 0; i < curdef->GetNumTriangles(); i++ )
      {
         scratch_cull_array[i] = CheckPlaneEquation( camera_position, curdef->GetTriangle( i ).surface_normal, curdef->GetTriangle( i ).plane_constant );
         if ( curdef->GetTriangle( i ).type == TtFlatShaded )
            scratch_intensity_array[i] = ComputeIntensity( light_source, curdef->GetTriangle( i ).surface_normal );
      }
   }
   else {
      for ( int i = 0; i < curdef->GetNumTriangles(); i++ )
      {
         if ( ( scratch_cull_array[i] = CheckPlaneEquation( camera_position, curdef->GetTriangle( i ).surface_normal, curdef->GetTriangle( i ).plane_constant ) ) == false )
         {
            scratch_intensity_array[i] = ComputeIntensity( light_source, curdef->GetTriangle( i ).surface_normal );
         }
      }
   }
}

/*
** bool AstObject::DetermineVisibility( void )
**
** This routine determines if an object is visible or not to the
** camera.
*/
bool AstObject::DetermineVisibility( void )
{

  /*
  ** See if object is entirely behind the viewer.  We do this
  ** by getting the object's position in the WCS and
  ** transforming the WCS value into a CCS value via a
  ** vector/matrix multiply.  After that, we then check its Z
  ** location with its bounding sphere
  */
  location_CCS = GetPositionWCS() * *world_to_camera_matrix;
  if ( location_CCS.Z + bounding_sphere < near_z )
  {
     SetFlag( ofVisible, false );
  }
  else
  {
     /*
     ** If the object is entirely in front of the viewer, we
     ** must then check its projected bounding sphere
     */
     if ( CheckProjectedBoundingSphere() )
     {
        /*
        ** We then select a representation (i.e. object
        ** definition) based on our CCS Z value.
        */
        if ( SelectRepresentation() )
           SetFlag( ofVisible, true );
        else
           SetFlag( ofVisible, false );
     }
     else
     {
        SetFlag( ofVisible, false );
     }
  }
  return CheckFlag( ofVisible );
}

/*
** void AstObject::Draw()
**
** This is the core drawing routine for an object.  It is
** responsible for doing the following steps:
**
**   1.  Cull and shade.
**   2.  Transforming points from its LCS to the CCS
**   3.  Projecting the resulting point.
**   5.  Drawing the resulting triangles
*/
void AstObject::Draw( void )
{
   CullAndShade();
   TransformVerticesToCCS();
   ProjectVertices();
   DrawTriangles();
}

/*
** inline unsigned char CalculateColor( intensity, color_base )
**
** This computes a VGA palette index (8-bit) based on a fixed
** point intensity (0..1) and a color_base (0...15).  The
** following equation is used:
**
**    palette index = intensity * 15 + color_base * 16
*/
inline unsigned char
CalculateColor( const AstFixedpoint_t intensity, const int color_base )
{
int intensity_value  = AstConvertToInt( AstFixMulFix( AstConvertToFix( 15 ), intensity ) );
int color_base_value = color_base * 16;

   return ( unsigned char ) ( intensity_value + color_base_value );
}

/*
** void AstObject::DrawTriangles( void )
**
** This routine is responsible for the actual rendering of the visible
** triangles to the screen.  Note that there are two different
** loops depending on whether texture mapping is enabled or not.
*/
void AstObject::DrawTriangles( void )
{
   if ( config.texturing )
   {
      for ( int i = 0; i < curdef->GetNumTriangles(); i++ )
      {
         if ( !scratch_cull_array[i] )
         {
            switch ( curdef->GetTriangle( i ).type )
            {
               case TtFlatShaded:
                  vga->Triangle( scratch_screen_array[curdef->GetTriangle( i ).vertices[0]],
                                 scratch_screen_array[curdef->GetTriangle( i ).vertices[1]],
                                 scratch_screen_array[curdef->GetTriangle( i ).vertices[2]],
                                 CalculateColor( scratch_intensity_array[i], curdef->GetTriangle( i ).color ) );
                  break;
               case TtTextureMapped:
                  vga->TextureMappedTriangle( scratch_screen_array[curdef->GetTriangle( i ).vertices[0]],
                                              scratch_screen_array[curdef->GetTriangle( i ).vertices[1]],
                                              scratch_screen_array[curdef->GetTriangle( i ).vertices[2]],
                                              curdef->GetTriangle( i ).texture_vertices[0],
                                              curdef->GetTriangle( i ).texture_vertices[1],
                                              curdef->GetTriangle( i ).texture_vertices[2],
                                              curdef->GetTriangle( i ).map );
                  break;
            }
         }
      }
   }
   else {
      for ( int i = 0; i < curdef->GetNumTriangles(); i++ )
      {
         if ( !scratch_cull_array[i] )
         {
            vga->Triangle( scratch_screen_array[curdef->GetTriangle( i ).vertices[0]],
                           scratch_screen_array[curdef->GetTriangle( i ).vertices[1]],
                           scratch_screen_array[curdef->GetTriangle( i ).vertices[2]],
                           CalculateColor( scratch_intensity_array[i], curdef->GetTriangle( i ).color ) );
         }
      }
   }
}

/*
** void AstObject::FreeScratchArrays( void )
**
** This routine is the converse of AllocateScratchArrays(), freeing
** up memory allocated by that routine.
*/
void AstObject::FreeScratchArrays( void )
{
   if ( scratch_vertex_array ) {
      delete []scratch_vertex_array;
      scratch_vertex_array = 0;
   }
   if ( scratch_screen_array ) {
      delete []scratch_screen_array;
      scratch_screen_array = 0;
   }
   if ( scratch_cull_array ) {
      delete []scratch_cull_array;
      scratch_cull_array = 0;
   }
   if ( scratch_intensity_array ) {
      delete []scratch_intensity_array;
      scratch_intensity_array = 0;
   }
}

/*
** bool AstObject::Intersects( const AstObject *object )
**
** This routine is used by the collision detector in the AstWorld
** class to determine if the bounding spheres of two objects
** intersect.  This is handled by comparing the distance
** between two objects with the bounding spheres of both of
** those objects.
*/
bool AstObject::Intersects( const AstObject *object ) const
{
AstVector between = GetPositionWCS() - object->GetPositionWCS();
long d = between.MagnitudeSquared();

   if ( d < object->GetBoundingSphereSquared() ||
        d < GetBoundingSphereSquared() )
   {
      return true;
   }
   return false;
}

/*
** the following routines handle translation by creating a
** 4x3 translation matrix and applying it to the local_matrix.
*/
static AstMatrix4x3 translation_matrix;

void AstObject::MoveAlongX( AstFixedpoint_t amount )
{
   translation_matrix.MakeIdentity();
   translation_matrix.MakeXTranslation( amount );
   local_matrix = translation_matrix * local_matrix;
}

void AstObject::MoveAlongY( AstFixedpoint_t amount )
{
   translation_matrix.MakeIdentity();
   translation_matrix.MakeYTranslation( amount );
   local_matrix = translation_matrix * local_matrix;
}

void AstObject::MoveAlongZ( AstFixedpoint_t amount )
{
   translation_matrix.MakeIdentity();
   translation_matrix.MakeZTranslation( amount );
   local_matrix = translation_matrix * local_matrix;
}

/*
** void AstObject::ProjectVertices( void )
**
** As the name implies, this routine projects the vertices in the
** object.  The vertices are retrieved from the scratch_vertex_array,
** which contains CCS data, and projected (using the camera, which
** contains field of view information) into the scratch_screen_array.
*/
void AstObject::ProjectVertices( void )
{
   camera.ProjectPoints( scratch_vertex_array, scratch_screen_array, curdef->GetNumVertices() );
}

/*
** The following functions handle rotations.  Unlike translation,
** I have opted not to use straight matrix multiplies to implement
** rotation.  The reason for this is one of accuracy -- because of
** the limited accuracy of fixed point math (or floating point
** math, for that matter) matrix values tend to start skewing
** after many multiplies.  This may take a while to build up, but
** when it does the effects can be quite dramatic -- heavily
** distorted objects on screen.
**
** To get around this we "fix" our matrices using an orthogonalization
** method whereby we compute the individual right, up, and forward
** vectors of the rotation matrix through cross products.  If this
** seems complicated -- don't worry about it, it works.
*/
static AstMatrix4x3 rotation_matrix;

void AstObject::RotateAroundX( unsigned short tenths_of_degrees )
{
AstFixedpoint_t m10 = local_matrix[1][0],
                m20 = local_matrix[2][0],
                m11 = local_matrix[1][1],
                m21 = local_matrix[2][1];

   rotation_matrix.MakeXRotation( tenths_of_degrees );

AstFixedpoint_t sm11 = rotation_matrix[1][1],
                sm12 = rotation_matrix[1][2],
                sm21 = rotation_matrix[2][1],
                sm22 = rotation_matrix[2][2];
   /*
   ** Compute column 0
   */
   local_matrix[1][0] = AstFixMulFix( sm11, m10 ) +
                        AstFixMulFix( sm12, m20 );
   local_matrix[2][0] = AstFixMulFix( sm21, m10 ) +
                        AstFixMulFix( sm22, m20 );

   /*
   ** Compute column 1
   */
   local_matrix[1][1] = AstFixMulFix( sm11, m11 ) +
                        AstFixMulFix( sm12, m21 );
   local_matrix[2][1] = AstFixMulFix( sm21, m11 ) +
                        AstFixMulFix( sm22, m21 );

   /*
   ** Compute column 2 (as col0 x col1 )
   */
   local_matrix[0][2] = AstFixMulFix( local_matrix[1][0], local_matrix[2][1] ) -
                        AstFixMulFix( local_matrix[2][0], local_matrix[1][1] );
   local_matrix[1][2] = AstFixMulFix( local_matrix[2][0], local_matrix[0][1] ) -
                        AstFixMulFix( local_matrix[0][0], local_matrix[2][1] );
   local_matrix[2][2] = AstFixMulFix( local_matrix[0][0], local_matrix[1][1] ) -
                        AstFixMulFix( local_matrix[0][1], local_matrix[1][0] );

   /*
   ** Recompute column 0 as col1 x col2
   */
   AstVector vu, vf, new_vr;

   local_matrix.GetVF( vf );
   local_matrix.GetVU( vu );
   new_vr = vu.Cross( vf );
   local_matrix.SetVR( new_vr );

   local_matrix.NormalizeDirectionVectors();
}
void AstObject::RotateAroundY( unsigned short tenths_of_degrees )
{
AstFixedpoint_t m00 = local_matrix[0][0],
                m20 = local_matrix[2][0],
                m01 = local_matrix[0][1],
                m21 = local_matrix[2][1];

   rotation_matrix.MakeYRotation( tenths_of_degrees );

AstFixedpoint_t sm00 = rotation_matrix[0][0],
                sm02 = rotation_matrix[0][2],
                sm20 = rotation_matrix[2][0],
                sm22 = rotation_matrix[2][2];

   /*
   ** Compute column 0
   */
   local_matrix[0][0] = AstFixMulFix( sm00, m00 ) +
                        AstFixMulFix( sm02, m20 );
   local_matrix[2][0] = AstFixMulFix( sm20, m00 ) +
                        AstFixMulFix( sm22, m20 );

   /*
   ** Compute column 1
   */
   local_matrix[0][1] = AstFixMulFix( sm00, m01 ) +
                        AstFixMulFix( sm02, m21 );
   local_matrix[2][1] = AstFixMulFix( sm20, m01 ) +
                        AstFixMulFix( sm22, m21 );

   /*
   ** Compute column 2 (as col0 x col1 )
   */
   local_matrix[0][2] = AstFixMulFix( local_matrix[1][0], local_matrix[2][1] ) -
                        AstFixMulFix( local_matrix[2][0], local_matrix[1][1] );
   local_matrix[1][2] = AstFixMulFix( local_matrix[2][0], local_matrix[0][1] ) -
                        AstFixMulFix( local_matrix[0][0], local_matrix[2][1] );
   local_matrix[2][2] = AstFixMulFix( local_matrix[0][0], local_matrix[1][1] ) -
                        AstFixMulFix( local_matrix[0][1], local_matrix[1][0] );

   /*
   ** Recompute column 0 (as col1 x col2 )
   */
   AstVector vu, vf, new_vr;

   local_matrix.GetVF( vf );
   local_matrix.GetVU( vu );
   new_vr = vu.Cross( vf );
   local_matrix.SetVR( new_vr );

   local_matrix.NormalizeDirectionVectors();
}
void AstObject::RotateAroundZ( unsigned short tenths_of_degrees )
{
AstFixedpoint_t m00 = local_matrix[0][0],
                m01 = local_matrix[0][1],
                m10 = local_matrix[1][0],
                m11 = local_matrix[1][1];

   rotation_matrix.MakeZRotation( tenths_of_degrees );

AstFixedpoint_t sm00 = rotation_matrix[0][0],
                sm01 = rotation_matrix[0][1],
                sm10 = rotation_matrix[1][0],
                sm11 = rotation_matrix[1][1];

   /*
   ** Compute column 0
   */
   local_matrix[0][0] = AstFixMulFix( sm00, m00 ) +
                        AstFixMulFix( sm01, m10 );

   local_matrix[1][0] = AstFixMulFix( sm10, m00 ) +
                        AstFixMulFix( sm11, m10 );

   /*
   ** Compute column 1
   */
   local_matrix[0][1] = AstFixMulFix( sm00, m01 ) +
                        AstFixMulFix( sm01, m11 );
   local_matrix[1][1] = AstFixMulFix( sm10, m01 ) +
                        AstFixMulFix( sm11, m11 );

   /*
   ** Compute column 2 (as col0 x col1 )
   */
   local_matrix[0][2] = AstFixMulFix( local_matrix[1][0], local_matrix[2][1] ) -
                        AstFixMulFix( local_matrix[2][0], local_matrix[1][1] );
   local_matrix[1][2] = AstFixMulFix( local_matrix[2][0], local_matrix[0][1] ) -
                        AstFixMulFix( local_matrix[0][0], local_matrix[2][1] );
   local_matrix[2][2] = AstFixMulFix( local_matrix[0][0], local_matrix[1][1] ) -
                        AstFixMulFix( local_matrix[0][1], local_matrix[1][0] );


   /*
   ** Recompute column 0 (as col1 x col2 )
   */
   AstVector vu, vf, new_vr;

   local_matrix.GetVF( vf );
   local_matrix.GetVU( vu );
   new_vr = vu.Cross( vf );
   local_matrix.SetVR( new_vr );

   local_matrix.NormalizeDirectionVectors();
}

/*
** bool AstObject::SelectRepresentation( void )
**
** This routine selects the appropriate object definition to use
** based on the object's distance (Z-wise) from the camera.  If
** the object is "out of range" then "false" is returned and
** "curdef" is set to NULL.
*/
bool AstObject::SelectRepresentation( void )
{
   if ( location_CCS.Z < fardef_Z )
   {
      if ( location_CCS.Z < neardef_Z )
         curdef = neardef;
      else
         curdef = fardef;
      return true;
   }
   else
   {
      curdef = 0;
      return false;
   }
}

/*
** void AstObject::TransformVerticesToCCS( void )
**
** This member function is used to transform an object's
** vertices, defined in the LCS, into the CCS.  This implies
** an intermediate transformation into the LCS, which is
** accomplished via the concatenation of two separate
** matrices, the world_to_camera_matrix and the local_to_world
** matrix.  We take the composite matrix and apply it against
** all of our object definition's vertices, then store the
** transformed vertices into our scratch_vertex_array.
*/
void AstObject::TransformVerticesToCCS( void )
{
static AstMatrix4x3 lcs2ccs_matrix;

   /*
   ** Create a composite LCS to CCS matrix that consists
   ** of the product of the local_matrix and the
   ** global world_to_camera matrix.
   */
   lcs2ccs_matrix = local_matrix * *world_to_camera_matrix;

   /*
   ** Iterate through all of the object's vertices and transform
   ** them into the CCS, storing them in scratch_vertex_array.
   */
   for ( int i = 0; i < curdef->GetNumVertices(); i++ ) {
      scratch_vertex_array[i] = curdef->GetVertex( i ) * lcs2ccs_matrix;
   }
}
