#include "ast3d.h"

/*
** This is used internally by the AstWorld class as an upper
** bound on the number of pointers that its visible_objects
** array can store.  The current value is higher than the
** total number of objects AST3D contains, but I prefer having
** some slack.
*/
const int AstWorld::max_visible_objects = 100;

/*
** This is a global pointer world_to_object transformation matrix
** for the object that is acting as the camera.  This is then
** used for local-to-camera transformations by all of the other
** objects in the world during rendering.
*/
const AstMatrix4x3 *world_to_camera_matrix;

/*
** AstWorld::AstWorld( void )
**
** This is the AstWorld's constructor.  It initializes various
** variables, allocates memory for the array of visible_objects,
** and sets up the light source.
*/
AstWorld::AstWorld( void )
{
   object_list_head = 0;

   if ( ( visible_objects = new AstObject *[max_visible_objects] ) == 0 )
      AstFatal( "Could not allocate visible_objects!" );

   AstVector ls( AstConvertToFix( 1 ),
                 AstConvertToFix( 1 ),
                 AstConvertToFix( 1 ) );
   ls.Normalize();
   SetLightSource( ls );
}

/*
** AstWorld::~AstWorld()
**
** This is the AstWorld class' destructor.  It deallocates any
** memory allocated by the AstWorld class.
*/
AstWorld::~AstWorld()
{
   delete []visible_objects;
   visible_objects = 0;
}

/*
** void AstWorld::Add( AstObject &o )
**
** This routine is responsible for inserting objects into the
** world.  Objects are stored in a linked list, the head of
** which is contained in the object_list_head pointer.  This
** routine also traps and disallows multiple inclusions of the
** same object into the world.
*/
void AstWorld::Add( AstObject &o )
{
   /*
   ** If no objects have been inserted into the world yet
   ** then make this object the head of the object list.
   */
   if ( object_list_head == 0 ) {
      object_list_head = &o;
      o.next_in_world = 0;
   }
   /*
   ** Otherwise find the last object in the world and then
   ** set the passed object as the new last object in the
   ** world.
   */
   else {
      AstObject *last_in_world = object_list_head;

      /*
      ** Find the last object in the world
      */
      while ( last_in_world->next_in_world != 0 ) {

         /*
         ** See if the passed object has already been inserted
         */
         if ( last_in_world == &o )
            return;
         last_in_world = last_in_world->next_in_world;
      }

      /*
      ** Adjust pointers
      */
      last_in_world->next_in_world = &o;
      o.next_in_world = 0;
   }
}

/*
** void AstWorld::Run( void )
**
** This is the core program loop for AST3D -- it automates the
** entire program by updating variables, input devices, and
** objects; checking for collisions; then rendering the objects
** within the world.
*/
void AstWorld::Run( void )
{
   /*
   ** This state variable is set to true when AstWorld::Stop()
   ** is called (the responsibility of the player).
   */
   while ( !state.done )
   {
      /*
      ** Update input devices
      */
      if ( gameport )
         gameport->Update();
      if ( inputdevice2d )
         inputdevice2d->Update();

      /*
      ** Do the rest of the stuff
      */
      CheckCollisions();
      UpdateObjectList();
      DrawObjectList();
   }
}

/*
** void AstWorld::CalculateVisibilities( void )
**
** This routine is the basic wrapper function for the AstWorld
** class responsible for checking visibilities.  It simply iterates
** through the linked list of objects in the world, informing each
** of them to determine its respective visibility.  Pointers to any
** visible objects are stored in an array of pointers known, oddly
** enough, as "visible_objects".
*/
void AstWorld::CalculateVisibilities( void )
{

   /*
   ** Reset count variable
   */
   num_visible_objects = 0;

   /*
   ** Iterate through the list of objects in the world.
   */
   AstObject *current_object = object_list_head;

   while ( current_object != 0 ) {

      /*
      ** Determine visibility, and if visible store the address
      ** of the object in the next available slow in the
      ** visible_objects array.
      */
      if ( current_object->DetermineVisibility() ) {
         if ( num_visible_objects < max_visible_objects )
            visible_objects[num_visible_objects++] = current_object;
         else
            AstFatal( "Too many visible objects!" );
      }
      current_object = current_object->next_in_world;
   }
}

/*
** void AstWorld::CheckCollisions( void )
**
** This function is responsible for iterating through the linked
** list of objects in the world and checking them for collisions.
** Collision detection is done via bounding spheres, which while
** not completely accurate, is much faster than a true object/object
** intersection test.
**
** When two objects are known to collide, their respective virtual
** "AstObject::YouHit( AstObject *object )" functions are called.
*/
void AstWorld::CheckCollisions( void )
{
AstObject *object = object_list_head;

   // Iterate through the list of objects in the world
   while ( object != 0 ) {
      // Only check against solid objects
      if ( object->CheckFlag( ofSolid ) ) {
         AstObject *object2 = object->next_in_world;

         // Iterate through the objects after the next object
         while ( object2 ) {
            // Make sure it's solid too
            if ( object2->CheckFlag( ofSolid ) ) {
               // If an intersection occurs
               if ( object->Intersects( object2 ) ) {
                  // Inform both
                  object->YouHit( object2 );
                  object2->YouHit( object );
               }
            }
            object2 = object2->next_in_world;
         }
      }
      object = object->next_in_world;
   }
}

/*
** void AstWorld::DrawCrosshairs( void )
**
** Simple routine to draw a crude set of crosshairs so the user
** has some idea where their input device is oriented.
*/
void AstWorld::DrawCrosshairs( void )
{
   if ( inputdevice2d && !config.joystick )
   {
     int sx, sy;

     sx = AstConvertToInt( AstFixMulFix( inputdevice2d->GetX(), AstConvertToFix( 160 ) ) ) + SCREEN_WIDTH/2;
     sy = AstConvertToInt( AstFixMulFix( inputdevice2d->GetY(), AstConvertToFix( 100 ) ) ) + SCREEN_HEIGHT/2;

     vga->HLine( sx - 10, sx + 10, sy, 150 );
     vga->VLine( sx, sy - 8, sy + 8, 150 );
   }
}

/*
** void AstWorld::DrawObjectList( void )
**
** This routine is responsible for drawing all of the objects in
** the world.  The basic pipeline is quite visible in the code
** itself.
*/
void AstWorld::DrawObjectList( void )
{

   /*
   ** Erase the framebuffer
   */
   vga->ClearScreen( 0 );

   /*
   ** Calculate our global world_to_camera_matrix so that we can
   ** transform object positions into the CCS.
   */
   world_to_camera_matrix = player->CalculateWorldToObjectMatrix();

   /*
   ** Let all of the objects in the world determine their
   ** visibility based on Z location and projected bounding
   ** sphere extents.  Visible objects are stored in the
   ** "visible_objects" array.
   */
   CalculateVisibilities();

   /*
   ** Sort the objects in the "visible_objects" array by depth.
   */
   SortObjectList();

   /*
   ** Draw the depth sorted objects
   */
   for ( int i = 0; i < num_visible_objects; i++ )
      visible_objects[i]->Draw();

   /*
   ** Give the user an indication of where the input device
   ** is pointing.
   */
   DrawCrosshairs();

   /*
   ** Then blit the page out.
   */
   vga->BlitPage();
}

/*
** The following routines all have to deal with depth sorting
** the objects in the world.
*/
typedef int (*qsortfnc)(const void *, const void *);

/*
** This is the sorting comparator function used by qsort().  It
** simply checks the Z values of the two passed objects and
** returns a comparison value based on their relation.
*/
int sortfnc( const AstObject **a, const AstObject **b )
{
register AstFixedpoint_t z1 = ( *a )->GetPositionCCS().Z,
                         z2 = ( *b )->GetPositionCCS().Z;

   if ( z1 < z2 )
      return 1;
   else if ( z1 > z2 )
      return -1;
   return 0;
}

/*
** void AstWorld::SortObjectList( void )
**
** This routine sorts the array of pointers to visible objects
** based on depth.  "sortfnc()" is used as the comparator, and
** the standard library qsort() function is used to do the
** actual sorting.  In practice, an application-specific sorting
** procedure and algorithm would be used, but in our situation
** qsort() works just fine.
*/
void AstWorld::SortObjectList( void )
{
   if ( num_visible_objects > 1 )
   {
      qsort( visible_objects,
             num_visible_objects,
             sizeof( AstObject * ),
             ( qsortfnc ) sortfnc );
   }
}

/*
** void AstWorld::UpdateObjectList( void )
**
** Updates all of the objects contained in the world by telling
** them to "DoSomething()".
*/
void AstWorld::UpdateObjectList( void )
{
   /*
   ** Inform all the objects in the world that they should update
   */
   AstObject *current_object = object_list_head;
   while ( current_object != 0 ) {
      current_object->DoSomething();
      current_object = current_object->next_in_world;
   }
}
